' PARSERWORDS.BAS
' Version 3

' HTML nested table parsing demo

' A Very Fast Configurable Parser (PB Assembler)
' with Test Demo
'
' Charles E V Pegge
' 2 July 2007
' added phrase tokenising ability  (specified with undersores between words)
' also
' matchid function for identifying words from short lists

' charles@pegge.net
' www.pegge.net
'
'
'
' with thanks to the members of the PowerBasic and the Jose Roca Peer Support Forums.

'
'

' Parsing rules using the default TokenType table:
'
' space, and all lower ascii and all non alphanumerics are treated as delimiters. eg:
' But an underscore _ (ascii 95) is treated as an alphanumeric eg:
'
'   FREE.LUNCH is interpreted as three words
'   FREE_LUNCH is one word

' all other non alphanumerics including '-' (minus) are self delimiting eg:
'
'   + - . ,  { } are always recognised as individual words even when unspaced
'
'
' But quoted text can be treated as an entire word including the outer quotes.
' inner quotes can be embedded eg:
'
'    "She said 'HI' "
'    'She said "HI" '
'
' In this implementation, this quoted text skipping is turned on and off dynamically during the 
' process, according to the context.
'
' quick Ascii code reference:
' ---------------------------
'  various control chars 0-31:
'  null      0
'  htab      9
'  linefeed  10
'  vtab      11
'  formfeed  12
'  creturn   13
'  escape    27
'  space     32
'  !"#$%&'   33-39
'  ()*+,-./  40-47
'  0-9       48-57
'  :;<=>?@   58-64
'  A-Z       65-90
'  [\]^_`    91-96
'  a-z       97-122
'  {|}~      123-126
' ---------------------------



#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

MACRO TWORD=MID$(dstring,TokenStart, TokenEnd-TokenStart)
MACRO ENABLE_QUOTE_SKIPPING=@tblp[34]=33: @tblp[39]=33
MACRO DISABLE_QUOTE_SKIPPING=@tblp[34]=35: @tblp[39]=35
MACRO GET_TOKEN=v=GetNextToken(STRPTR(dString), lString, TokenStart, TokenEnd)


#INCLUDE "ParserWords.inc" ' Low level funvtions in Assembler




FUNCTION WINMAIN (BYVAL hInstance AS DWORD,BYVAL hPrevInstance AS DWORD,BYVAL lpCmdLine AS ASCIIZ PTR,BYVAL iCmdShow AS LONG) AS LONG


  %MaxEntries=2000  ' an arbitrary figure
'

'============================================'
'  CREATING THEN TOKENTYPE AND TOKEN TABLES  '
'============================================'



DIM cmz AS STRING
'sample Reserved Word list.
'
' any word found on this list will return
' the corresponding 2 byte hexadecimal code
' instead of the TokenType Code from the table below.
'

' Keyword List Rules:
'---------------------

' The token must always be a 4 character hexadecimal number.
' There must be only one space (or tab) between the HexCode and the keyword.
' Keywords/phrases  must alphanumeric or underscore maximum of 255 characters.
'
' The keywords are not case sensitive.
' Apart from the single space between code and keyword, a minimum of 1 spacing 
' character is used to separate the data items.
' Spacing characters are any from ascii 0 to ascii 32 inclusive, including
' CRs, LFs TABs etc.
'
' Rules for Phrases:
'
' Phrases must be listed before their first individual word.
' The words in a phrase are connected by a single underscore.
' A phrase does not mind multiple spaces or line breaks between text words.
' Underscores in the text are accepted by the phrase matching process.
' 
' During the search process, it does not matter how many spaces, linefeeds or tabs 
' are between the words.
' 
' EXAMPLE:
' 03AB IN_THE_STYLE_OF 
' placed in the token table will identify  any of these, returning &h03AB :
'
' In the     style of
' in    the style of
' in [tab] the style [linefeed] of
' in_the_style_of
' in-the-style-of
' in_the_style-of
' 
' It will NOT identify:

' in_ the style_ of
' in__the style of
' in- the style of
' _in the style of

'
' Note that words that come first on the list have priority over later ones,
' so if a phrase begins with IN then the phrase IN_THE_STYLE must come before it,
' and the phrase: IN_THE_STYLE_OF must come even before this.
'
'
' In this example, HTML tag words have tokens of form: 01xx for
' easy identigication by the parsing loop.
' Other words to be identified map from 02xx onward.
'
' Some HTML keywords and a phrase for this demo:
'
cmz=_
" 0100 html"+_
" 0101 head"+_
" 0102 body"+_
" 0103 div"+_
" 0104 table"+_
" 0105 tr"+_
" 0106 td"+_
" 0107 a"+_
" 0108 img"+_
" 010A script"+_
" 010B style"+_
" 0200 in_the_style_of"+_
" 0201 in"+_
" 0202 the"+_
" 0203 this"+_
" "

' TokenType Lookup Table

 DIM tbl  AS GLOBAL STRING:   tbl=NUL$(&h2100+LEN(cmz)+%MaxEntries)
 DIM tblp AS GLOBAL BYTE PTR: tblp=STRPTR(tbl)

 '  The TokenType Table:  '

 ' BY DEFAULT, QUOTES (&h22 " and &h27 ' ) ARE CONFIGURED AS ORDINARY SYMBOLS
 ' AND QUOTE SKIPPING IS THEREFORE NOT ACTIVEATED.

 '0123456789ABCDEF'     '
 MID$(tbl,1)= _                 '
 "                " + _ '0
 "                " + _ '1
 " ###############" + _ '2
 "0000000000######" + _ '3
 "#AAAAAAAAAAAAAAA" + _ '4
 "AAAAAAAAAAA####_" + _ '5
 "#aaaaaaaaaaaaaaa" + _ '6
 "aaaaaaaaaaa#####" + _ '7
 "hhhhhhhhhhhhhhhh" + _ '8
 "hhhhhhhhhhhhhhhh" + _ '9
 "hhhhhhhhhhhhhhhh" + _ 'A
 "hhhhhhhhhhhhhhhh" + _ 'B
 "hhhhhhhhhhhhhhhh" + _ 'C
 "hhhhhhhhhhhhhhhh" + _ 'D
 "hhhhhhhhhhhhhhhh" + _ 'E
 "hhhhhhhhhhhhhhhh" + _ 'F
 ""

' Tokentype Codings used in table
'--------------------------------
' 20  space/ignore      32
' 21  quote marks     ! 33' for quote skipping
' 22  (uncommitted)   " 34
' 23  (uncommited)    # 35
' 24  (uncommitted)   $ 36
' 25  (uncommitted)   % 37
' 26  (uncommitted)   & 38
' 27  (uncommitted)   ' 29
' 28  (reserved)      ( 40
' 29  (reserved)      ) 41
' 2A  (uncommitted)   * 42
' 2B  (uncommitted)   + 43
' 2C  (uncommitted)   , 44
' 2D  (uncommitted)   - 45
' 2E  (uncommitted)   . 46
' 2F  (uncommitted)   / 47
' 30  numbers         0 48
' 41  alpha uppercase A 65
' 61  alpha lowercase a 97
' 68  HIGHER ASCII    h 104
'


'
' TokenType Table Rules
'
' These 4 rules are hard coded into the Assembler parsing routine:
'
' 1. any code less than 33 is treated as space.
' 2. code 33 signifies a quote, and all characters between and including the matching
'    quote marks will be included in the word.
' 3. symbol types (from 34 to 47) are understood to be self-delimiting.
' 4. any code  greater than 47 is treated as text (so that includes the underscore 95).
'
'
' Configurability
'
' within these rules you can devise your own codings.
' When the GetNextToken is called, it will return the coding associated with the first character of the word.
' It is possible to generate codes for 'return' (13), and 'tab' (9) or any other lower ascii
' if your lexicon requires it. 
'
' You can designate any character to be a symbol, even a linefeed. In this program, we assign line feeds 
' to return a $ (36) 
'
' You can designate any character to be a quote marker as well as the usual " (34) and ' (39), by
' placing a ! (33) in the table at the marker's ascii position.
'
'
' TABLE OVERRIDES  '

'  Patch specifics into  TokenType table  '
' 
' this can be done dynamically to alter the parsing rules contextually.

' activate these 2 in scripts, and between tag markers
' set them back to 35 for main text
' @tblp[34]=33 ' 34 " tokentype 33  for quote skipping
' @tblp[34]=33 ' 39 " tokentype 33  for quote skipping

@tblp[10]=36 '    $ tokentype 36  for identifying linefeeds
@tblp[60]=40 ' 60 < tokentype 40  for detecting tags
@tblp[62]=41 ' 62 > tokentype 41  for detecting tag endings.
@tblp[47]=47 ' 47 / tokentype 47  for detecting end-tags.



DIM errp AS LONG, ii AS LONG, jj AS LONG ' to report errors during table compilation

ReadInWords(cmz)

' this function returns several items of information on the CPU registers.
' whereas a normal function only returns things in its eax register.
' These items can be stored directly from the registers into LONG variables,
' immediately following the function call.

! mov errp,eax
! mov ii,edx
! mov jj,ecx

' This enables error reporting, if anything went wrong reading the reserved words.

IF errp THEN
 errp=MAX(errp-STRPTR(cmz)-8,1)
 MSGBOX "Problem with tokenlist about here: "+MID$(cmz,errp,16) +$CR _
 + HEX$(STRPTR(cmz))+" "+HEX$(ii)+" ecx: "+HEX$(jj)
 EXIT FUNCTION
END IF

'
DIM dString    AS STRING              ' the text to be read
DIM lString    AS LONG                ' length of this text
DIM pString    AS BYTE PTR            ' actual address of the start of text
DIM TokenStart AS LONG: TokenStart=1  ' string index to the 1st character of a word. (this is an output-only value)
DIM TokenEnd  AS LONG                 ' string index to the position immediately following the word.
                                      ' (set this to wherever you want to begin)

DIM s      AS STRING                  ' string to receive the report to be displayed
DIM i      AS LONG                    ' iterator
DIM j      AS LONG                    ' iterator
DIM v      AS LONG                    ' token
DIM w      AS LONG                    ' token
DIM exr    AS LONG                    ' loop drop out flag
DIM TblNo  AS LONG                    ' table counter
DIM tdepth AS LONG                    ' nested tables depth
DIM trow   AS LONG                    ' nested table row number
DIM tcol   AS LONG                    ' nested tablescolumn number
DIM tbli   AS LONG                    ' subtable pointer index
DIM tblj   AS LONG                    ' subtable pointer index
DIM tblp(1023,4) AS STATIC LONG       ' subTable pointers for recursive action
'          '





'==============--'
' Decoding HTML  '
'================'

' HTML has quite a complex syntax which changes according to whether you are inside or
' outside a tag or whether you are in a comment or embedded Javascript. This demo shows how to 
' get through the HTML script safely and read the data you are interested in.
'
' An additional complication is nested tables for which a recursive procedure is 
' required to separate out the individual tables and their contents, and to provide
' mapping references for one table to another.
'
' The most challenging aspect of writing this programme was to give a simple
' and well ordered expression to a very convoluted process.


'====================='
'  The sample script  '
'====================='

' As a web page this sample is utter gibberish, but it tests the programme quite well.


 ERRCLEAR
 OPEN "somehtml.htm" FOR INPUT AS 1:CLOSE 1
 IF ERR THEN MSGBOX "File 'somehtml.htm' Not Accessible": EXIT FUNCTION
 OPEN "somehtml.htm" FOR BINARY ACCESS READ AS 1:GET$ 1,LOF(1),dString:CLOSE 1


lString=LEN(dString)
pString=STRPTR(dString)  ' address of the string content
                         '
                         ' first a dummy loop to measure the PowerBasic overheads
                         '


s="Tests: "+$CR


'===================='
' MAIN PARSING LOOP  '
'===================='

DO ' for both main and nested tables

 DO  ' main lexical loop
  GET_TOKEN
  SELECT CASE v ' identify primary token
  CASE 0
   EXIT LOOP ' no more text
  'case 41
  CASE 36
   v=-1 ' line feeds ignore
  CASE 40 ' for tag starts <
   ENABLE_QUOTE_SKIPPING
   GET_TOKEN
   IF (v AND &hff00)=&h100 THEN
    GOSUB HTML_tag_types
   ELSEIF v=35 THEN  ' generalsymbols
    IF ASC(dstring,TokenStart)=33 THEN GOSUB skip_comment ' <!
   ELSEIF v=41 THEN ' >
    s=s+"<>"
    DISABLE_QUOTE_SKIPPING
   ELSEIF v=47 THEN ' / termination tag
    '----------------------------------
    GET_TOKEN
    IF v=41 THEN ' /
     ' non specific </>
     DISABLE_QUOTE_SKIPPING
     EXIT SELECT
    ELSEIF v=&h104 THEN ' </table>
     IF tdepth<1 THEN GOSUB skip_tag_attributes: exr=1: EXIT DO
     IF tdepth<>0 THEN
      IF exr=1 THEN s=s+"End Nested Table."+STR$(TblNo)+$CR: EXIT DO
     END IF
     DECR tdepth:trow=0:tcol=0
     IF tdepth=0 THEN s=s+"End Table."+STR$(TblNo)+$CR
    ELSEIF v=&h105 THEN ' </tr>
     tcol=0: s=s+$CR
     IF tdepth<1 THEN exr=1: EXIT DO 'malformed html
    ELSEIF v=&h106 THEN  ' </td>
     s=s+$CR
     IF tdepth<1 THEN exr=1:EXIT DO 'mal formed html
    END IF
    GOSUB skip_tag_attributes
   END IF ' token type 47
   '-------------------------------------
  CASE ELSE ' unidentified primary tokens
   IF tdepth>0 THEN ' show only if within current table, excludes data outside a table.
    IF (v>&hff) THEN ' identified phrases text outside tables
     s=s+"["+HEX$(v)+"]"+TWORD+" "
    END IF
    IF (v>0)AND(v<&hff) THEN ' all text outside top <td>s
     s=s+TWORD+" "
    END IF
   END IF
  END SELECT ' primary token

 LOOP ' next primary token  until end of text

'-------------------------'
'  RESOLVE TESTED TABLES  '
'-------------------------'

IF tbli>tblj THEN
 INCR TblNo
 s=s+$CR+"Nested Table "+STR$(TblNo) _
  +" depth:"+STR$(tblp(tblj,1)) _
  +"  in Table "+STR$(tblp(tblj,4)) _
  +" row "+STR$(tblp(tblj,2)) _
  +"  col "+STR$(tblp(tblj,3))+$CR
 '
 TokenEnd=tblp(tblj):INCR tblj
 trow=0:tcol=0:tdepth=1:exr=1
ELSE
 EXIT LOOP
END IF

LOOP

MSGBOX s


'--------------------------------'

EXIT FUNCTION


'---------------'
'  SUBROUTINES  '
'---------------'

'  SUBROUTINES FOR TAG_TYPES  '


HTML_tag_types:

 SELECT CASE v ' identify html tag types
 CASE &h101 'head
  GOSUB skip_head
 CASE &h104 'table
  IF tdepth>0 THEN
   GOSUB skip_tables
  ELSEIF tdepth=0 THEN
   INCR TblNo: s=s+$CR+"Table:"+STR$(TblNo)+$CR
   INCR tdepth: trow=0:tcol=0
   GOSUB skip_tag_attributes
  END IF
 CASE &h105 'tr
  IF tdepth=1 THEN INCR trow: s=s+"Row:"+STR$(trow)+$CR
  GOSUB skip_tag_attributes
 CASE &h106 'td
  IF tdepth=1 THEN INCR tcol: s=s+"Column:"+STR$(tcol)+$CR
  GOSUB skip_tag_attributes
 CASE &h10A
  GOSUB skip_script
 CASE ELSE ' ignore all other tags
  GOSUB skip_tag_attributes
 END SELECT ' html tag type
RETURN





' SUBROUTINES FOR SKIPPING  '



skip_tag_attributes:

 DO
  GET_TOKEN
  IF (v=0)OR(v=41) THEN
   v=-1
   DISABLE_QUOTE_SKIPPING
   EXIT DO
  END IF
 LOOP
RETURN


skip_head:
 GOSUB skip_tag_attributes
 DO
  GET_TOKEN
  IF v=0 THEN EXIT DO
  IF v=40 THEN
   ENABLE_QUOTE_SKIPPING
   GET_TOKEN
   IF v=47 THEN
    GET_TOKEN
    IF v=&h101 THEN ' </head>
     GOSUB skip_tag_attributes: EXIT DO
    END IF
    GOSUB skip_tag_attributes
   END IF
  END IF
 LOOP
 v=-1
RETURN




skip_tables:  ' RECURSIVE SUBROUTINE

 INCR tdepth
 GOSUB skip_tag_attributes
 ' you may want to log the position of the table
 ' here for later reading
 IF exr=0 THEN
  tblp(tbli,0)=TokenEnd
  tblp(tbli,1)=tdepth
  tblp(tbli,2)=trow
  tblp(tbli,3)=tcol
  tblp(tbli,4)=TblNo
  INCR tbli
  ' limits
  IF tbli>1023 THEN
   exr=-1
   MSGBOX "Warning!"+$CR+"Nested Table Log is full for this web page."+$CR+"Cannot log any more of them here."
  END IF
 END IF
 DO
  GET_TOKEN
  IF v=0 THEN EXIT DO
  IF v=40 THEN
   ENABLE_QUOTE_SKIPPING
   GET_TOKEN
   IF v=&h104 THEN
    GOSUB skip_tables: v=-1 ' --- recurse! ----
   END IF
   IF v=47 THEN
    GET_TOKEN
    IF v=&h104 THEN ' </table>
     DECR tdepth
     GOSUB skip_tag_attributes: EXIT DO
    END IF
   END IF
   GOSUB skip_tag_attributes
  END IF
 LOOP
 v=-1
RETURN



skip_script:

 GOSUB skip_tag_attributes
 ENABLE_QUOTE_SKIPPING
 DO
  GET_TOKEN
  IF v=0 THEN EXIT DO
  IF v=40 THEN
   GET_TOKEN
   IF v=47 THEN
    GET_TOKEN
    IF v=&h10A THEN ' </script>
     GOSUB skip_tag_attributes: EXIT DO
    END IF
   END IF
  END IF
 LOOP
 v=-1
RETURN




skip_comment:

 ' <!
 DO
  GET_TOKEN
  IF v=0 THEN EXIT DO
  IF v=35 THEN ' a symbol
   w=ASC(dstring,TokenStart)
   IF w=45 THEN
    GET_TOKEN
    w=ASC(dstring,TokenStart)
    IF w=45 THEN
     GET_TOKEN
     IF v=41 THEN ' -->
      DISABLE_QUOTE_SKIPPING
      EXIT DO
     END IF
    END IF
   END IF
  END IF
 LOOP
 v=-1
RETURN


END FUNCTION


'=============='
' END PROGRAM  '
'=============='

